# -*- coding: utf-8 -*-
"""
The Blackboard Integrated Channel package.
"""

__version__ = "0.0.1"

default_app_config = ("integrated_channels.blackboard.apps."
                      "BlackboardConfig")



from django.db import migrations

ADDITIONAL_ENTERPRISE_ENROLLMENT_SOURCES = {
    'Customer Admin Enrollment': 'customer_admin'
}

def add_new_enterprise_enrollment_source(apps, schema_editor):
    enrollment_sources = apps.get_model('enterprise', 'EnterpriseEnrollmentSource')
    for name, slug in ADDITIONAL_ENTERPRISE_ENROLLMENT_SOURCES.items():
        enrollment_sources.objects.update_or_create(name=name, slug=slug)


def drop_new_enterprise_enrollment_source(apps, schema_editor):
    enrollment_sources = apps.get_model('enterprise', 'EnterpriseEnrollmentSource')
    enrollment_sources.objects.filter(name__in=ADDITIONAL_ENTERPRISE_ENROLLMENT_SOURCES).delete()

class Migration(migrations.Migration):

    dependencies = [
        ('enterprise', '0139_auto_20210803_1854'),
    ]

    operations = [
        migrations.RunPython(
            code=add_new_enterprise_enrollment_source, reverse_code=drop_new_enterprise_enrollment_source
        )

    ]



# -*- coding: utf-8 -*-
"""
Utilities for Cornerstone integrated channels.
"""

import re
from logging import getLogger
from uuid import uuid4

from django.apps import apps

from integrated_channels.utils import encode_course_key_into_base64


def cornerstone_learner_data_transmission_audit():
    """
    Returns the ``EnterpriseCustomer`` class.
    """
    return apps.get_model('cornerstone', 'CornerstoneLearnerDataTransmissionAudit')


def cornerstone_course_key_model():
    """
    Returns the ``EnterpriseCustomer`` class.
    """
    return apps.get_model('cornerstone', 'CornerstoneCourseKey')


LOGGER = getLogger(__name__)


def create_cornerstone_learner_data(request, course_id):
    """
        updates or creates CornerstoneLearnerDataTransmissionAudit
    """
    try:
        defaults = {
            'user_guid': request.GET['userGuid'],
            'session_token': request.GET['sessionToken'],
            'callback_url': request.GET['callbackUrl'],
            'subdomain': request.GET['subdomain'],
        }
        cornerstone_learner_data_transmission_audit().objects.update_or_create(
            user_id=request.user.id,
            course_id=course_id,
            defaults=defaults
        )
    except KeyError:
        # if we couldn't find a key, it means we don't want to save data. just skip it by doing nothing.
        pass
    except Exception as ex:  # pylint: disable=broad-except
        LOGGER.error('Unable to Create/Update CornerstoneLearnerDataTransmissionAudit. {ex}'.format(ex=ex))


def convert_invalid_course_id(course_id):
    """
    Regex check a course ID to see if it contains any invalid chars. If it does then encode the string, otherwise
    return the original course ID.
    """
    safe_course_id = course_id
    re2 = re.compile(r"[|<>.&%\s\\/\“]+")
    if re2.search(safe_course_id):
        # If the course key contains any of the invalid chars, encode the key
        safe_course_id = encode_course_key_into_base64(course_id)
    # If the encoded or unencoded version of the key are over 50 characters, they will error out
    # in cornerstone, so we convert them to a uuid.
    if len(safe_course_id) > 50:
        safe_course_id = str(uuid4())
    return safe_course_id


def get_or_create_key_pair(course_id):
    key_mapping, ___ = cornerstone_course_key_model().objects.get_or_create(
        internal_course_id=course_id, defaults={
            'external_course_id': convert_invalid_course_id(course_id)})
    return key_mapping



# Generated by Django 2.2.15 on 2020-08-06 17:56

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise', '0107_remove_branding_config_banner_fields'),
    ]

    operations = [
        migrations.AddField(
            model_name='historicallicensedenterprisecourseenrollment',
            name='is_revoked',
            field=models.BooleanField(default=False, help_text="Whether the licensed enterprise course enrollment is revoked, e.g., when a user's license is revoked."),
        ),
        migrations.AddField(
            model_name='licensedenterprisecourseenrollment',
            name='is_revoked',
            field=models.BooleanField(default=False, help_text="Whether the licensed enterprise course enrollment is revoked, e.g., when a user's license is revoked."),
        ),
    ]



# -*- coding: utf-8 -*-
"""
Class for transmitting content metadata to SuccessFactors.
"""

import logging
from itertools import islice

from django.conf import settings

from integrated_channels.exceptions import ClientError
from integrated_channels.integrated_channel.transmitters.content_metadata import ContentMetadataTransmitter
from integrated_channels.sap_success_factors.client import SAPSuccessFactorsAPIClient
from integrated_channels.utils import chunks, generate_formatted_log

LOGGER = logging.getLogger(__name__)


class SapSuccessFactorsContentMetadataTransmitter(ContentMetadataTransmitter):
    """
    This transmitter transmits exported content metadata to SAPSF.
    """

    def __init__(self, enterprise_configuration, client=SAPSuccessFactorsAPIClient):
        """
        Use the ``SAPSuccessFactorsAPIClient`` for content metadata transmission to SAPSF.
        """
        super().__init__(
            enterprise_configuration=enterprise_configuration,
            client=client
        )

    def transmit(self, create_payload, update_payload, delete_payload, content_updated_mapping, **kwargs):
        """
        Transmit content metadata items to the integrated channel.
        """
        # Since delete_payload is a mapping between content key and past transmission record, we need to create a
        # mapping between content key and channel_metadata. Then we can prepare the channel metadata for deletion
        items_to_delete = {key: item.channel_metadata for key, item in delete_payload.items()}
        self._prepare_items_for_delete(items_to_delete)

        prepared_items = {}
        prepared_items.update(create_payload)

        # Do the same work to the update payload as was done to the delete payload above
        items_to_update = {key: item.channel_metadata for key, item in update_payload.items()}
        prepared_items.update(items_to_update)
        prepared_items.update(items_to_delete)

        chunk_items = chunks(prepared_items, self.enterprise_configuration.transmission_chunk_size)
        transmission_limit = settings.INTEGRATED_CHANNELS_API_CHUNK_TRANSMISSION_LIMIT.get(
            self.enterprise_configuration.channel_code()
        )
        for chunk in islice(chunk_items, transmission_limit):
            chunked_items = list(chunk.values())
            try:
                self.client.update_content_metadata(self._serialize_items(chunked_items))
            except ClientError as exc:
                LOGGER.error(
                    generate_formatted_log(
                        self.enterprise_configuration.channel_code(),
                        self.enterprise_configuration.enterprise_customer.uuid,
                        None,
                        None,
                        f'Failed to update [{len(chunked_items)}] content metadata items for '
                        f'integrated channel {self.enterprise_configuration.enterprise_customer.name} '
                        f'{self.enterprise_configuration.channel_code()}'
                    )
                )
                LOGGER.exception(exc)

                # Remove the failed items from the create/update/delete dictionaries,
                # so ContentMetadataItemTransmission objects are not synchronized for
                # these items below.
                self._remove_failed_items(chunked_items, create_payload, items_to_update, items_to_delete)

        # If API transmission limit is set then mark the rest of the items as not transmitted.
        # Since, chunk_items is a generator and we have already iterated through the items that need to
        # be transmitted. Rest of the items are the ones that need to marked as not transmitted.
        for chunk in chunk_items:
            chunked_items = list(chunk.values())
            self._remove_failed_items(chunked_items, create_payload, items_to_update, items_to_delete)

        self._create_transmissions(create_payload, content_updated_mapping)
        self._update_transmissions(update_payload, content_updated_mapping)
        self._delete_transmissions(list(items_to_delete.keys()))

    def _remove_failed_items(self, failed_items, items_to_create, items_to_update, items_to_delete):
        """
        Remove content metadata items from the `items_to_create`, `items_to_update`, `items_to_delete` dicts.

        Arguments:
            failed_items (list): Failed Items to be removed.
            items_to_create (dict): dict containing the items created successfully.
            items_to_update (dict): dict containing the items updated successfully.
            items_to_delete (dict): dict containing the items deleted successfully.
        """
        for item in failed_items:
            content_metadata_id = item['courseID']
            items_to_create.pop(content_metadata_id, None)
            items_to_update.pop(content_metadata_id, None)
            items_to_delete.pop(content_metadata_id, None)

    def _prepare_items_for_transmission(self, channel_metadata_items):
        return {
            'ocnCourses': channel_metadata_items
        }

    def _prepare_items_for_delete(self, items_to_delete):
        """
        Set status to INACTIVE for items that should be deleted.
        """
        for item in items_to_delete.values():
            item['status'] = 'INACTIVE'



# -*- coding: utf-8 -*-
"""
URL definitions for Cornerstone API.
"""

from django.conf.urls import url

from integrated_channels.cornerstone.views import CornerstoneCoursesListView, CornerstoneCoursesUpdates

urlpatterns = [
    url(
        r'^course-list$',
        CornerstoneCoursesListView.as_view(),
        name='cornerstone-course-list'
    ),
    url(
        r'course-updates',
        CornerstoneCoursesUpdates.as_view(),
        name='cornerstone-course-updates'
    )
]



# -*- coding: utf-8 -*-
"""
Client for connecting to Canvas.
"""
import json
import logging
from http import HTTPStatus

import requests
from dateutil.parser import parse
from six.moves.urllib.parse import quote_plus, urljoin

from django.apps import apps

from integrated_channels.canvas.utils import CanvasUtil
from integrated_channels.exceptions import ClientError
from integrated_channels.integrated_channel.client import IntegratedChannelApiClient
from integrated_channels.utils import generate_formatted_log, refresh_session_if_expired

LOGGER = logging.getLogger(__name__)


MESSAGE_WHEN_COURSE_WAS_DELETED = 'Course was deleted previously, skipping create/update'


class CanvasAPIClient(IntegratedChannelApiClient):
    """
    Client for connecting to Canvas.

    Required Canvas auth credentials to instantiate a new client object.
        -  canvas_base_url : the base url of the user's Canvas instance.
        -  client_id : the ID associated with a Canvas developer key.
        -  client_secret : the secret key associated with a Canvas developer key.
        -  refresh_token : the refresh token token retrieved by the `oauth/complete`
        endpoint after the user authorizes the use of their Canvas account.

    Order of Operations:
        Before the client can connect with an Enterprise user's Canvas account, the user will need to
        follow these steps
            - Create a developer key with their Canvas account
            - Provide the ECS team with their developer key's client ID and secret.
            - ECS will return a url for the user to visit which will prompt authorization and redirect when
            the user hits the `confirm` button.
            - The redirect will hit the `oauth/complete` endpoint which will use the passed oauth code
            to request the Canvas oauth refresh token and save it to the Enterprise user's Canvas API config
            - The refresh token is used at client instantiation to request the user's access token, this access
            token is saved to the client's session and is used to make POST and DELETE requests to Canvas.
    """

    def __init__(self, enterprise_configuration):
        """
        Instantiate a new client.

        Args:
            enterprise_configuration (CanvasEnterpriseCustomerConfiguration): An enterprise customers's
            configuration model for connecting with Canvas
        """
        super().__init__(enterprise_configuration)
        self.config = apps.get_app_config('canvas')
        self.session = None
        self.expires_at = None
        self.course_create_url = CanvasUtil.course_create_endpoint(self.enterprise_configuration)

    def create_content_metadata(self, serialized_data):
        """
        Creates a course in Canvas.
        If course is not found, easy!  create it as usual
        If course is found, it will have one of the following `workflow_state` values:
                available: issue an update with latest field values
                completed: this happens if a course has been concluded. Update it to change status
                  to offer by using course[event]=offer (which makes course published in Canvas)
                unpublished: still just update
                deleted: take no action for now.
        For information of Canvas workflow_states see `course[event]` at:
        https://canvas.instructure.com/doc/api/courses.html#method.courses.update
        """
        self._create_session()

        desired_payload = json.loads(serialized_data.decode('utf-8'))
        course_details = desired_payload['course']

        edx_course_id = course_details['integration_id']
        located_course = CanvasUtil.find_course_by_course_id(
            self.enterprise_configuration,
            self.session,
            edx_course_id
        )

        if not located_course:
            LOGGER.info(
                generate_formatted_log(
                    'canvas',
                    self.enterprise_configuration.enterprise_customer.uuid,
                    None,
                    edx_course_id,
                    f'Creating new course with payload {desired_payload}',
                )
            )
            # Course does not exist: Create the course
            status_code, response_text = self._post(
                self.course_create_url,
                serialized_data,
            )
            created_course_id = json.loads(response_text)['id']

            # step 2: upload image_url and any other details
            self._update_course_details(created_course_id, course_details)
        else:
            workflow_state = located_course['workflow_state']
            if workflow_state.lower() == 'deleted':
                LOGGER.error(
                    generate_formatted_log(
                        'canvas',
                        self.enterprise_configuration.enterprise_customer.uuid,
                        None,
                        edx_course_id,
                        'Course with integration_id = {edx_course_id} found in deleted state, '
                        'not attempting to create/update'.format(
                            edx_course_id=edx_course_id,
                        ),
                    )
                )
                status_code = 200
                response_text = MESSAGE_WHEN_COURSE_WAS_DELETED
            else:
                # 'unpublished', 'completed' or 'available' cases
                LOGGER.warning(
                    generate_formatted_log(
                        'canvas',
                        self.enterprise_configuration.enterprise_customer.uuid,
                        None,
                        edx_course_id,
                        'Course with canvas_id = {course_id},'
                        'integration_id = {edx_course_id} found in workflow_state={workflow_state},'
                        ' attempting to update instead of creating it'.format(
                            course_id=located_course["id"],
                            edx_course_id=edx_course_id,
                            workflow_state=workflow_state,
                        ),
                    )
                )
                status_code, response_text = self._update_course_details(
                    located_course['id'],
                    course_details,
                )

        return status_code, response_text

    def update_content_metadata(self, serialized_data):
        self._create_session()

        integration_id = self._extract_integration_id(serialized_data)
        course_id = CanvasUtil.get_course_id_from_edx_course_id(
            self.enterprise_configuration,
            self.session,
            integration_id,
        )

        url = CanvasUtil.course_update_endpoint(
            self.enterprise_configuration,
            course_id,
        )

        return self._put(url, serialized_data)

    def delete_content_metadata(self, serialized_data):
        self._create_session()

        integration_id = self._extract_integration_id(serialized_data)
        course_id = CanvasUtil.get_course_id_from_edx_course_id(
            self.enterprise_configuration,
            self.session,
            integration_id,
        )

        url = '{}/api/v1/courses/{}'.format(
            self.enterprise_configuration.canvas_base_url,
            course_id,
        )

        return self._delete(url)

    def create_assessment_reporting(self, user_id, payload):
        """
        Send assessment level learner data, retrieved by the integrated channels exporter, to Canvas in the form of
        an assignment and submission.
        """
        learner_data = json.loads(payload)
        self._create_session()

        # Retrieve the Canvas user ID from the user's edx email (it is assumed that the learner's Edx
        # and Canvas emails will match).
        canvas_user_id = self._search_for_canvas_user_by_email(user_id)

        canvas_course_id = self._handle_get_user_canvas_course(canvas_user_id, learner_data['courseID'])

        # Depending on if the assignment already exists, either retrieve or create it.
        # Assessment level reporting Canvas assignments use the subsection ID as the primary identifier, whereas
        # course level reporting assignments rely on the course run key.
        assignment_id = self._handle_canvas_assignment_retrieval(
            learner_data['subsectionID'],
            canvas_course_id,
            learner_data['subsection_name'],
            learner_data['points_possible'],
            is_assessment_grade=True
        )

        # The percent grade from the grades api is represented as a decimal, but we can report the percent in the
        # request body as the string: `<int percent grade>%`
        update_grade_response = self._handle_canvas_assignment_submission(
            "{}%".format(str(learner_data['grade'] * 100)),
            canvas_course_id,
            assignment_id,
            canvas_user_id
        )

        return update_grade_response.status_code, update_grade_response.text

    def create_course_completion(self, user_id, payload):
        learner_data = json.loads(payload)
        self._create_session()

        # Retrieve the Canvas user ID from the user's edx email (it is assumed that the learner's Edx
        # and Canvas emails will match).
        canvas_user_id = self._search_for_canvas_user_by_email(user_id)

        canvas_course_id = self._handle_get_user_canvas_course(canvas_user_id, learner_data['courseID'])

        # Depending on if the assignment already exists, either retrieve or create it.
        assignment_id = self._handle_canvas_assignment_retrieval(
            learner_data['courseID'],
            canvas_course_id,
            '(Edx integration) Final Grade'
        )

        # Course completion percentage grades are exported as decimals but reported to Canvas as integer percents.
        update_grade_response = self._handle_canvas_assignment_submission(
            learner_data['grade'] * 100,
            canvas_course_id,
            assignment_id,
            canvas_user_id
        )

        return update_grade_response.status_code, update_grade_response.text

    def delete_course_completion(self, user_id, payload):
        # Todo: There isn't a great way for users to delete course completion data
        pass

    def cleanup_duplicate_assignment_records(self, courses):
        """
        For each course provided, iterate over assessments contained within the associated Canvas course and remove all
        but the most recent, unique assessments sorted by `updated_at`.

        Args:
            - courses: iterable set of unique course IDs
        """
        self._create_session()
        failures = []
        num_assignments_removed = 0
        num_failed_assignments = 0
        for edx_course in courses:
            canvas_course = CanvasUtil.find_course_by_course_id(
                self.enterprise_configuration,
                self.session,
                edx_course
            )

            # Add any missing courses to a list of failed courses
            if not canvas_course:
                failures.append(edx_course)
                continue

            canvas_assignments_url = CanvasUtil.course_assignments_endpoint(
                self.enterprise_configuration,
                canvas_course['id']
            )

            # Dict of most current, unique assignments in the course
            current_assignments = {}

            # Running list of duplicate assignments (ID's) that need to be deleted
            assignments_to_delete = []

            current_page_count = 0
            more_pages_present = True

            # Continue iterating over assignment responses while more paginated results exist or until the page count
            # limit is hit
            while more_pages_present and current_page_count < 150:
                resp = self.session.get(canvas_assignments_url)

                if resp.status_code >= 400:
                    LOGGER.error(
                        generate_formatted_log(
                            'canvas',
                            self.enterprise_configuration.enterprise_customer.uuid,
                            None,
                            edx_course,
                            'Failed to retrieve assignments for Canvas course: {} while running deduplication, '
                            'associated edx course: {}'.format(
                                canvas_course['id'],
                                edx_course
                            )
                        )
                    )
                    more_pages_present = False
                else:
                    # Result of paginated response from the Canvas course assignments API
                    assignments_resp = resp.json()

                    # Ingest Canvas assignments API response and replace older duplicated assignments in current
                    # assignments. All older duplicated assignment IDs are added to `assignments_to_delete`
                    current_assignments, assignments_to_delete = self._parse_unique_newest_assignments(
                        current_assignments,
                        assignments_to_delete,
                        assignments_resp
                    )

                    # Determine if another page of results exists
                    next_page = CanvasUtil.determine_next_results_page(resp)
                    if next_page:
                        canvas_assignments_url = next_page
                        current_page_count += 1
                    else:
                        more_pages_present = False

            # Remove all assignments from the current course and record the number of assignments removed
            assignments_removed, individual_assignment_failures = \
                self._bulk_remove_course_assignments(
                    canvas_course.get('id'),
                    assignments_to_delete
                )
            num_assignments_removed += len(assignments_removed)
            num_failed_assignments += len(individual_assignment_failures)

        if failures or num_failed_assignments:
            message = 'Failed to dedup all assignments for the following courses: {}. ' \
                      'Number of individual assignments that failed to be deleted: {}. ' \
                      'Total assignments removed: {}.'.format(
                          failures,
                          num_failed_assignments,
                          num_assignments_removed
                      )
            status_code = 400
        else:
            message = 'Removed {} duplicate assignments from Canvas.'.format(num_assignments_removed)
            status_code = 200

        return status_code, message

    def update_participation_types(self, canvas_pks):
        """
        For each canvas course provided, send an update with the parameter of
        'restrict_enrollments_to_course_dates' to true in order to set the participation types of
        all canvas courses to "Course" instead of "Term", which allows users to view the end date

        Args:
            - canvas_pk: list of primary keys of previously transmitted courses from canvas customers
        """
        self._create_session()

        for course_pk in canvas_pks:
            integration_id = course_pk['content_id']
            try:
                course_id = CanvasUtil.get_course_id_from_edx_course_id(
                    self.enterprise_configuration,
                    self.session,
                    integration_id,
                )

                update_payload = {'course[restrict_enrollments_to_course_dates]': True}
                url = CanvasUtil.course_update_endpoint(
                    self.enterprise_configuration,
                    course_id,
                )
                self._put(url, json.dumps(update_payload).encode('utf-8'))
            except ClientError:
                LOGGER.info(
                    generate_formatted_log(
                        'canvas',
                        self.enterprise_configuration.enterprise_customer.uuid,
                        None,
                        integration_id,
                        f'Skipped course with id {integration_id}, not found in Canvas',
                    )
                )

    # Private Methods
    def _bulk_remove_course_assignments(self, course_id, assignments_to_remove):
        """
        Take a Canvas course ID and remove all assessments associated a list of Canvas course assignment IDs.

        Args:
            - course_id: Canvas course ID
            - assignments_to_remove: List of assignment ID's to be removed contained with the provided course.
        """
        removed_items = []
        failures = []
        for assignment_id in assignments_to_remove:
            try:
                assignment_url = CanvasUtil.course_assignments_endpoint(
                    self.enterprise_configuration,
                    course_id
                ) + '/{}'.format(assignment_id)
                self._delete(assignment_url)
                removed_items.append(assignment_id)
            except ClientError:
                # we do not want assignment deletes to cause failures
                failures.append(assignment_id)
        return removed_items, failures

    def _parse_unique_newest_assignments(self, current_assignments, assignments_to_delete, assignment_response_json):
        """
        Ingest an assignments response from Canvas into a dictionary of most current, unique assignments found and a
        running list of assignments to delete

        Args:
            - current_assignments: dictionary containing information on most current unique assignments contained within
            a Canvas course.
                Example:
                    {
                        'edX+816': {
                            'id': 10,
                            'updated_at': '2021-06-10T13:57:19Z',
                        },
                        'edX+100': {
                            'id': 11,
                            'updated_at': '2021-06-10T13:58:19Z',
                        }
                    }

            - assignments_to_delete: list of Canvas assignment IDs associated with duplicate assignments to be deleted

            - assignment_response_json: json repr of the requests' Response object returned by Canvas' course
            assignments API
        """
        for assignment in assignment_response_json:
            integration_id = assignment['integration_id']
            current_assignment = current_assignments.get(integration_id)
            if current_assignment:
                if parse(current_assignment['updated_at']) < parse(assignment['updated_at']):
                    assignments_to_delete.append(current_assignment['id'])
                    current_assignments[integration_id] = {
                        'id': assignment['id'],
                        'updated_at': assignment['updated_at']
                    }
                else:
                    assignments_to_delete.append(assignment['id'])
            else:
                current_assignments[integration_id] = {
                    'id': assignment['id'],
                    'updated_at': assignment['updated_at']
                }

        return current_assignments, assignments_to_delete

    def _update_course_details(self, course_id, course_details):
        """
        Update a course for image_url (and possibly other settings in future).
        Also sets course to 'offer' state by sending 'course[event]=offer',
        which makes the course published in Canvas.

        Arguments:
          - course_id (Number): Canvas Course id
          - course_details (dict): { 'image_url' } : optional, used if present for course[image_url]
        """
        response_code = None
        response_text = None
        url = CanvasUtil.course_update_endpoint(
            self.enterprise_configuration,
            course_id,
        )
        # Providing the param `event` and setting it to `offer` is equivalent to publishing the course.
        update_payload = {'course': {'event': 'offer'}}
        try:
            # there is no way to do this in a single request during create
            # https://canvas.instructure.com/doc/api/all_resources.html#method.courses.update
            if "image_url" in course_details:
                update_payload['course']['image_url'] = course_details['image_url']

            response_code, response_text = self._put(url, json.dumps(update_payload).encode('utf-8'))
        except Exception as course_exc:  # pylint: disable=broad-except
            # we do not want course image update to cause failures
            edx_course_id = course_details["integration_id"]
            exc_string = str(course_exc)
            LOGGER.error(
                generate_formatted_log(
                    'canvas',
                    self.enterprise_configuration.enterprise_customer.uuid,
                    None,
                    edx_course_id,
                    'Failed to update details for course, '
                    'canvas_course_id={canvas_course_id}. '
                    'Details: {details}'.format(
                        canvas_course_id=course_id,
                        details=exc_string,
                    )
                )
            )

        return response_code, response_text

    def _post(self, url, data):
        """
        Make a POST request using the session object to a Canvas endpoint.

        Args:
            url (str): The url to send a POST request to.
            data (bytearray): The json encoded payload to POST.
        """
        post_response = self.session.post(url, data=data)
        if post_response.status_code >= 400:
            raise ClientError(post_response.text, post_response.status_code)
        return post_response.status_code, post_response.text

    def _put(self, url, data):
        """
        Make a PUT request using the session object to the Canvas course update endpoint

        Args:
            url (str): The canvas url to send update requests to.
            data (bytearray): The json encoded payload to UPDATE. This also contains the integration
            ID used to match a course with a course ID.
        """
        put_response = self.session.put(url, data=data)
        if put_response.status_code >= 400:
            raise ClientError(put_response.text, put_response.status_code)
        return put_response.status_code, put_response.text

    def _delete(self, url):
        """
        Make a DELETE request using the session object to the Canvas course delete endpoint.
        this actually only 'conclude's a course. See this link for difference between
        conclude and delete. Conclude allows bringing course back to 'offer' state
        https://canvas.instructure.com/doc/api/courses.html#method.courses.destroy

        Args:
            url (str): The canvas url to send delete requests to.
        """
        delete_response = self.session.delete(url, data='{"event":"conclude"}')
        if delete_response.status_code >= 400:
            raise ClientError(delete_response.text, delete_response.status_code)
        return delete_response.status_code, delete_response.text

    def _extract_integration_id(self, data):
        """
        Retrieve the integration ID string from the encoded transmission data and apply appropriate
        error handling.

        Args:
            data (bytearray): The json encoded payload intended for a Canvas endpoint.
        """
        if not data:
            raise ClientError("No data to transmit.", HTTPStatus.NOT_FOUND.value)

        try:
            decoded_payload = data.decode("utf-8")
            decoded_json = json.loads(decoded_payload)
        except AttributeError as error:
            raise ClientError(
                f"Unable to decode data. Type of data was {type(data)}", HTTPStatus.BAD_REQUEST.value
            ) from error

        try:
            integration_id = decoded_json['course']['integration_id']
        except KeyError as error:
            LOGGER.exception(generate_formatted_log(
                'canvas',
                self.enterprise_configuration.enterprise_customer.uuid,
                None,
                None,
                f'KeyError processing decoded json. decoded payload was: {decoded_payload}'
            ), exc_info=error)
            raise ClientError(
                "Could not transmit data, no integration ID present.", HTTPStatus.NOT_FOUND.value
            ) from error

        return integration_id

    def _search_for_canvas_user_by_email(self, user_email):
        """
        Helper method to make an api call to Canvas using the user's email as a search term.

        Args:
            user_email (string) : The email associated with both the user's Edx account and Canvas account.
        """
        get_user_id_from_email_url = '{url_base}/api/v1/accounts/{account_id}/users?search_term={email_address}'.format(
            url_base=self.enterprise_configuration.canvas_base_url,
            account_id=self.enterprise_configuration.canvas_account_id,
            email_address=quote_plus(user_email)  # emails with unique symbols such as `+` can cause issues
        )
        rsps = self.session.get(get_user_id_from_email_url)

        if rsps.status_code >= 400:
            raise ClientError(
                "Failed to retrieve user from Canvas: received response-[{}]".format(rsps.reason),
                rsps.status_code
            )

        get_users_by_email_response = rsps.json()

        try:
            canvas_user_id = get_users_by_email_response[0]['id']
        except (KeyError, IndexError) as error:
            raise ClientError(
                "No Canvas user ID found associated with email: {}".format(user_email),
                HTTPStatus.NOT_FOUND.value
            ) from error
        return canvas_user_id

    def _get_canvas_user_courses_by_id(self, user_id):
        """Helper method to retrieve all courses that a Canvas user is enrolled in."""
        get_users_courses_url = '{canvas_base_url}/api/v1/users/{canvas_user_id}/courses'.format(
            canvas_base_url=self.enterprise_configuration.canvas_base_url,
            canvas_user_id=user_id
        )
        rsps = self.session.get(get_users_courses_url)

        if rsps.status_code >= 400:
            raise ClientError(
                "Could not retrieve Canvas course list. Received exception: {}".format(
                    rsps.reason
                ),
                rsps.status_code
            )

        return rsps.json()

    def _handle_canvas_assignment_retrieval(
            self,
            integration_id,
            course_id,
            assignment_name,
            points_possible=100,
            is_assessment_grade=False
    ):
        """
        Helper method to handle course assignment creation or retrieval. Canvas requires an assignment
        in order for a user to get a grade, so first check the course for the "final grade"
        assignment. This assignment will have a matching integration id to the currently transmitting
        learner data. If this assignment is not yet created on Canvas, send a post request to do so.

        Args:
            integration_id (str) : the string integration id from the edx course.
            course_id (str) : the Canvas course ID relating to the course which the client is currently
            transmitting learner data to.
        """
        # Check if the course assignment already exists
        canvas_assignments_url = '{canvas_base_url}/api/v1/courses/{course_id}/assignments'.format(
            canvas_base_url=self.enterprise_configuration.canvas_base_url,
            course_id=course_id
        )
        resp = self.session.get(canvas_assignments_url)

        more_pages_present = True
        current_page_count = 0
        assignment_id = ''

        # current_page_count serves as a timeout, limiting to a max of 150 pages of requests
        while more_pages_present and current_page_count < 150:
            if resp.status_code >= 400:
                raise ClientError(
                    "Something went wrong retrieving assignments from Canvas. Got response: {}".format(
                        resp.text,
                    ),
                    resp.status_code
                )

            assignments_resp = resp.json()
            for assignment in assignments_resp:
                try:
                    if assignment['integration_id'] == integration_id:
                        assignment_id = assignment['id']
                        break

                # The integration ID check above should ensure that we have a 200 response from Canvas,
                # but sanity catch if we have a unexpected response format
                except (KeyError, ValueError, TypeError) as error:
                    raise ClientError(
                        "Something went wrong retrieving assignments from Canvas. Got response: {}".format(
                            resp.text,
                        ),
                        resp.status_code
                    ) from error

            if not assignment_id:
                next_page = CanvasUtil.determine_next_results_page(resp)
                if next_page:
                    resp = self.session.get(next_page)
                    current_page_count += 1
                else:
                    more_pages_present = False
            else:
                more_pages_present = False

        # Canvas requires a course assignment for a learner to be assigned a grade.
        # If no assignment has been made yet, create it.
        if not assignment_id:
            assignment_creation_data = {
                'assignment': {
                    'name': assignment_name,
                    'submission_types': 'none',
                    'integration_id': integration_id,
                    'published': True,
                    'points_possible': points_possible,
                    'omit_from_final_grade': is_assessment_grade,
                }
            }
            create_assignment_resp = self.session.post(canvas_assignments_url, json=assignment_creation_data)

            try:
                assignment_id = create_assignment_resp.json()['id']
            except (ValueError, KeyError) as error:
                raise ClientError(
                    "Something went wrong creating an assignment on Canvas. Got response: {}".format(
                        create_assignment_resp.text,
                    ),
                    create_assignment_resp.status_code
                ) from error
        return assignment_id

    def _handle_canvas_assignment_submission(self, grade, course_id, assignment_id, canvas_user_id):
        """
        Helper method to take necessary learner data and post to Canvas as a submission to the correlated assignment.
        """
        submission_url = '{base_url}/api/v1/courses/{course_id}/assignments/' \
            '{assignment_id}/submissions/{user_id}'.format(
                base_url=self.enterprise_configuration.canvas_base_url,
                course_id=course_id,
                assignment_id=assignment_id,
                user_id=canvas_user_id
            )

        # The percent grade from the grades api is represented as a decimal
        submission_data = {
            'submission': {
                'posted_grade': grade
            }
        }
        submission_response = self.session.put(submission_url, json=submission_data)

        if submission_response.status_code >= 400:
            raise ClientError(
                "Something went wrong while posting a submission to Canvas assignment: {} under Canvas course: {}."
                " Recieved response {} with the status code: {}".format(
                    assignment_id,
                    course_id,
                    submission_response.text,
                    submission_response.status_code
                )
            )
        return submission_response

    def _handle_get_user_canvas_course(self, canvas_user_id, learner_data_course_id):
        """
        Helper method to take the Canvas user ID and edX course ID to find the matching Canvas course information.
        """
        # With the Canvas user ID, retrieve all courses for the user.
        user_courses = self._get_canvas_user_courses_by_id(canvas_user_id)

        # Find the course who's integration ID matches the learner data course ID. This integration ID can be either
        # an edX course run ID or course ID. Raise if no course found.
        canvas_course_id = None
        for course in user_courses:
            if course['integration_id'] == learner_data_course_id:
                canvas_course_id = course['id']
                break

        if not canvas_course_id:
            raise ClientError(
                "Course: {course_id} not found registered in Canvas for Canvas learner: {canvas_user_id}.".format(
                    course_id=learner_data_course_id,
                    canvas_user_id=canvas_user_id,
                ),
                HTTPStatus.NOT_FOUND.value,
            )

        return canvas_course_id

    def _create_session(self):
        """
        Instantiate a new session object for use in connecting with Canvas. Each enterprise customer
        connecting to Canvas should have a single client session.
        Will only create a new session if token expiry has been reached
        """
        self.session, self.expires_at = refresh_session_if_expired(
            self._get_oauth_access_token,
            self.session,
            self.expires_at,
        )

    def _get_oauth_access_token(self):
        """Uses the client id, secret and refresh token to request the user's auth token from Canvas.

        Returns:
            access_token (str): the OAuth access token to access the Canvas API as the user
            expires_in (int): the number of seconds after which token will expire
        Raises:
            HTTPError: If we received a failure response code from Canvas.
            ClientError: If an unexpected response format was received that we could not parse.
        """
        client_id = self.enterprise_configuration.client_id
        client_secret = self.enterprise_configuration.client_secret

        if not client_id:
            raise ClientError(
                "Failed to generate oauth access token: Client ID required.",
                HTTPStatus.INTERNAL_SERVER_ERROR
            )
        if not client_secret:
            raise ClientError(
                "Failed to generate oauth access token: Client secret required.",
                HTTPStatus.INTERNAL_SERVER_ERROR
            )
        if not self.enterprise_configuration.refresh_token:
            raise ClientError(
                "Failed to generate oauth access token: Refresh token required.",
                HTTPStatus.INTERNAL_SERVER_ERROR
            )

        if not self.enterprise_configuration.canvas_base_url or not self.config.oauth_token_auth_path:
            raise ClientError(
                "Failed to generate oauth access token: Canvas oauth path missing from configuration.",
                HTTPStatus.INTERNAL_SERVER_ERROR
            )
        auth_token_url = urljoin(
            self.enterprise_configuration.canvas_base_url,
            self.config.oauth_token_auth_path,
        )

        auth_token_params = {
            'grant_type': 'refresh_token',
            'client_id': client_id,
            'client_secret': client_secret,
            'state': str(self.enterprise_configuration.enterprise_customer.uuid),
            'refresh_token': self.enterprise_configuration.refresh_token,
        }

        auth_response = requests.post(auth_token_url, auth_token_params)
        if auth_response.status_code >= 400:
            raise ClientError(auth_response.text, auth_response.status_code)
        try:
            data = auth_response.json()
            return data['access_token'], data["expires_in"]
        except (KeyError, ValueError) as error:
            raise ClientError(auth_response.text, auth_response.status_code) from error



# Generated by Django 3.2.8 on 2021-11-01 17:08

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import model_utils.fields
import simple_history.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('enterprise', '0145_auto_20211013_1018'),
    ]

    operations = [
        migrations.CreateModel(
            name='Degreed2LearnerDataTransmissionAudit',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('degreed_user_email', models.CharField(help_text='Used as the user-id field when creating a completion', max_length=255)),
                ('enterprise_course_enrollment_id', models.PositiveIntegerField(db_index=True)),
                ('course_id', models.CharField(help_text='Used as content-id field when creating a completion', max_length=255)),
                ('completed_timestamp', models.CharField(help_text='yyyy-mm-ddTHH:MM:SS format', max_length=19)),
                ('status', models.CharField(max_length=100)),
                ('error_message', models.TextField(blank=True)),
                ('created', models.DateTimeField(auto_now_add=True)),
            ],
        ),
        migrations.CreateModel(
            name='HistoricalDegreed2EnterpriseCustomerConfiguration',
            fields=[
                ('id', models.IntegerField(auto_created=True, blank=True, db_index=True, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('idp_id', models.CharField(blank=True, help_text='If provided, will be used as IDP slug to locate remote id for learners', max_length=255, null=True)),
                ('active', models.BooleanField(help_text='Is this configuration active?')),
                ('transmission_chunk_size', models.IntegerField(default=500, help_text='The maximum number of data items to transmit to the integrated channel with each request.')),
                ('channel_worker_username', models.CharField(blank=True, help_text='Enterprise channel worker username to get JWT tokens for authenticating LMS APIs.', max_length=255, null=True)),
                ('catalogs_to_transmit', models.TextField(blank=True, help_text='A comma-separated list of catalog UUIDs to transmit. If blank, all customer catalogs will be transmitted. If there are overlapping courses in the customer catalogs, the overlapping course metadata will be selected from the newest catalog.', null=True)),
                ('disable_learner_data_transmissions', models.BooleanField(default=False, help_text='When set to True, the configured customer will no longer receive learner data transmissions, both scheduled and signal based', verbose_name='Disable Learner Data Transmission')),
                ('key', models.CharField(help_text='The API Client ID provided to edX by the enterprise customer to be used to make API calls to Degreed on behalf of the customer.', max_length=255, verbose_name='API Client ID')),
                ('secret', models.CharField(help_text='The API Client Secret provided to edX by the enterprise customer to be used to make API calls to Degreed on behalf of the customer.', max_length=255, verbose_name='API Client Secret')),
                ('degreed_base_url', models.CharField(help_text='The base URL used for API requests to Degreed, i.e. https://degreed.com.', max_length=255, verbose_name='Degreed Base URL')),
                ('degreed_token_fetch_base_url', models.CharField(blank=True, help_text='If provided, will be used as base url instead of degreed_base_url to fetch tokens', max_length=255, null=True, verbose_name='Degreed token fetch base url')),
                ('history_id', models.AutoField(primary_key=True, serialize=False)),
                ('history_date', models.DateTimeField()),
                ('history_change_reason', models.CharField(max_length=100, null=True)),
                ('history_type', models.CharField(choices=[('+', 'Created'), ('~', 'Changed'), ('-', 'Deleted')], max_length=1)),
                ('enterprise_customer', models.ForeignKey(blank=True, db_constraint=False, help_text='Enterprise Customer associated with the configuration.', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='enterprise.enterprisecustomer')),
                ('history_user', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'historical degreed2 enterprise customer configuration',
                'ordering': ('-history_date', '-history_id'),
                'get_latest_by': 'history_date',
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name='Degreed2EnterpriseCustomerConfiguration',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('idp_id', models.CharField(blank=True, help_text='If provided, will be used as IDP slug to locate remote id for learners', max_length=255, null=True)),
                ('active', models.BooleanField(help_text='Is this configuration active?')),
                ('transmission_chunk_size', models.IntegerField(default=500, help_text='The maximum number of data items to transmit to the integrated channel with each request.')),
                ('channel_worker_username', models.CharField(blank=True, help_text='Enterprise channel worker username to get JWT tokens for authenticating LMS APIs.', max_length=255, null=True)),
                ('catalogs_to_transmit', models.TextField(blank=True, help_text='A comma-separated list of catalog UUIDs to transmit. If blank, all customer catalogs will be transmitted. If there are overlapping courses in the customer catalogs, the overlapping course metadata will be selected from the newest catalog.', null=True)),
                ('disable_learner_data_transmissions', models.BooleanField(default=False, help_text='When set to True, the configured customer will no longer receive learner data transmissions, both scheduled and signal based', verbose_name='Disable Learner Data Transmission')),
                ('key', models.CharField(help_text='The API Client ID provided to edX by the enterprise customer to be used to make API calls to Degreed on behalf of the customer.', max_length=255, verbose_name='API Client ID')),
                ('secret', models.CharField(help_text='The API Client Secret provided to edX by the enterprise customer to be used to make API calls to Degreed on behalf of the customer.', max_length=255, verbose_name='API Client Secret')),
                ('degreed_base_url', models.CharField(help_text='The base URL used for API requests to Degreed, i.e. https://degreed.com.', max_length=255, verbose_name='Degreed Base URL')),
                ('degreed_token_fetch_base_url', models.CharField(blank=True, help_text='If provided, will be used as base url instead of degreed_base_url to fetch tokens', max_length=255, null=True, verbose_name='Degreed token fetch base url')),
                ('enterprise_customer', models.OneToOneField(help_text='Enterprise Customer associated with the configuration.', on_delete=django.db.models.deletion.CASCADE, to='enterprise.enterprisecustomer')),
            ],
        ),
    ]



# -*- coding: utf-8 -*-
"""
Client for communicating with the Enterprise API.
"""

from collections import OrderedDict
from logging import getLogger

from django.conf import settings
from django.core.cache import cache

from enterprise import utils
from enterprise.api_client.lms import JwtLmsApiClient

LOGGER = getLogger(__name__)


class EnterpriseApiClient(JwtLmsApiClient):
    """
    Object builds an API client to make calls to the Enterprise API.
    """

    API_BASE_URL = settings.LMS_INTERNAL_ROOT_URL + '/enterprise/api/v1/'
    APPEND_SLASH = True

    ENTERPRISE_CUSTOMER_ENDPOINT = 'enterprise-customer'
    ENTERPRISE_CUSTOMER_CATALOGS_ENDPOINT = 'enterprise_catalogs'

    DEFAULT_VALUE_SAFEGUARD = object()

    def get_content_metadata(self, enterprise_customer, enterprise_catalogs=None):
        """
        Return all content metadata contained in the catalogs associated with the EnterpriseCustomer.

        Arguments:
            enterprise_customer (EnterpriseCustomer): The EnterpriseCustomer to return content metadata for.
            enterprise_catalogs (EnterpriseCustomerCatalog): Optional list of EnterpriseCustomerCatalog objects.

        Returns:
            list: List of dicts containing content metadata.
        """
        content_metadata = OrderedDict()
        enterprise_customer_catalogs = enterprise_catalogs or enterprise_customer.enterprise_customer_catalogs.all()
        for enterprise_customer_catalog in enterprise_customer_catalogs:
            response = self._load_data(
                self.ENTERPRISE_CUSTOMER_CATALOGS_ENDPOINT,
                resource_id=str(enterprise_customer_catalog.uuid),
                traverse_pagination=True,
                querystring={'page_size': 1000},
            )

            for item in response['results']:
                content_id = utils.get_content_metadata_item_id(item)
                content_metadata[content_id] = item

        return list(content_metadata.values())

    @JwtLmsApiClient.refresh_token
    def _load_data(
            self,
            resource,
            detail_resource=None,
            resource_id=None,
            querystring=None,
            traverse_pagination=False,
            default=DEFAULT_VALUE_SAFEGUARD,
    ):
        """
        Loads a response from a call to one of the Enterprise endpoints.

        :param resource: The endpoint resource name.
        :param detail_resource: The sub-resource to append to the path.
        :param resource_id: The resource ID for the specific detail to get from the endpoint.
        :param querystring: Optional query string parameters.
        :param traverse_pagination: Whether to traverse pagination or return paginated response.
        :param default: The default value to return in case of no response content.
        :return: Data returned by the API.
        """
        default_val = default if default != self.DEFAULT_VALUE_SAFEGUARD else {}
        querystring = querystring if querystring else {}

        cache_key = utils.get_cache_key(
            resource=resource,
            querystring=querystring,
            traverse_pagination=traverse_pagination,
            resource_id=resource_id
        )
        response = cache.get(cache_key)
        if not response:
            # Response is not cached, so make a call.
            endpoint = getattr(self.client, resource)(resource_id)
            endpoint = getattr(endpoint, detail_resource) if detail_resource else endpoint
            response = endpoint.get(**querystring)
            if traverse_pagination:
                results = utils.traverse_pagination(response, endpoint)
                response = {
                    'count': len(results),
                    'next': 'None',
                    'previous': 'None',
                    'results': results,
                }
            if response:
                # Now that we've got a response, cache it.
                cache.set(cache_key, response, settings.ENTERPRISE_API_CACHE_TIMEOUT)
        return response or default_val



# -*- coding: utf-8 -*-
"""
Database models for Enterprise Integrated Channel Degreed.
"""

from logging import getLogger

from simple_history.models import HistoricalRecords

from django.db import models

from integrated_channels.degreed2.exporters.content_metadata import Degreed2ContentMetadataExporter
from integrated_channels.degreed2.exporters.learner_data import Degreed2LearnerExporter
from integrated_channels.degreed2.transmitters.content_metadata import Degreed2ContentMetadataTransmitter
from integrated_channels.degreed2.transmitters.learner_data import Degreed2LearnerTransmitter
from integrated_channels.integrated_channel.models import EnterpriseCustomerPluginConfiguration

LOGGER = getLogger(__name__)


class Degreed2EnterpriseCustomerConfiguration(EnterpriseCustomerPluginConfiguration):
    """
    The Enterprise-specific configuration we need for integrating with Degreed2.

    .. no_pii:
    """

    client_id = models.CharField(
        max_length=255,
        verbose_name="API Client ID",
        help_text=(
            "The API Client ID provided to edX by the enterprise customer to be used to make API "
            "calls to Degreed on behalf of the customer."
        )
    )

    client_secret = models.CharField(
        max_length=255,
        verbose_name="API Client Secret",
        help_text=(
            "The API Client Secret provided to edX by the enterprise customer to be used to make API "
            "calls to Degreed on behalf of the customer."
        )
    )

    degreed_base_url = models.CharField(
        max_length=255,
        verbose_name="Degreed Base URL",
        help_text="The base URL used for API requests to Degreed, i.e. https://degreed.com."
    )

    degreed_token_fetch_base_url = models.CharField(
        max_length=255,
        null=True,
        blank=True,
        verbose_name="Degreed token fetch base url",
        help_text="If provided, will be used as base url instead of degreed_base_url to fetch tokens"
    )

    history = HistoricalRecords()

    class Meta:
        app_label = 'degreed2'

    def __str__(self):
        """
        Return human-readable string representation.
        """
        return "<Degreed2EnterpriseCustomerConfiguration for Enterprise {enterprise_name}>".format(
            enterprise_name=self.enterprise_customer.name
        )

    def __repr__(self):
        """
        Return uniquely identifying string representation.
        """
        return self.__str__()

    @staticmethod
    def channel_code():
        """
        Returns an capitalized identifier for this channel class, unique among subclasses.
        """
        return 'DEGREED2'

    def get_learner_data_transmitter(self):
        """
        Return a ``DegreedLearnerTransmitter`` instance.
        """
        return Degreed2LearnerTransmitter(self)

    def get_learner_data_exporter(self, user):
        """
        Return a ``DegreedLearnerDataExporter`` instance.
        """
        return Degreed2LearnerExporter(user, self)

    def get_content_metadata_transmitter(self):
        """
        Return a ``DegreedContentMetadataTransmitter`` instance.
        """
        return Degreed2ContentMetadataTransmitter(self)

    def get_content_metadata_exporter(self, user):
        """
        Return a ``DegreedContentMetadataExporter`` instance.
        """
        return Degreed2ContentMetadataExporter(user, self)


class Degreed2LearnerDataTransmissionAudit(models.Model):
    """
    The payload we sent to Degreed2 at a given point in time for an enterprise course enrollment.
    Ref: https://api.degreed.com/docs/#create-a-new-completion

    .. pii: The degreed_user_email model field contains PII.
    .. pii_types: email_address
    .. pii_retirement: consumer_api
    """

    degreed_user_email = models.CharField(
        max_length=255,
        blank=False,
        null=False,
        help_text='Used as the user-id field when creating a completion',
    )

    enterprise_course_enrollment_id = models.PositiveIntegerField(
        blank=False,
        null=False,
        db_index=True
    )

    course_id = models.CharField(
        max_length=255,
        blank=False,
        null=False,
        help_text="Used as content-id field when creating a completion"
    )

    completed_timestamp = models.CharField(
        max_length=19,
        help_text='yyyy-mm-ddTHH:MM:SS format',
    )

    # Request-related information.
    status = models.CharField(max_length=100, blank=False, null=False)
    error_message = models.TextField(blank=True)
    created = models.DateTimeField(auto_now_add=True)

    class Meta:
        app_label = 'degreed2'

    def __str__(self):
        """
        Return a human-readable string representation of the object.
        """
        return (
            '<Degreed2LearnerDataTransmissionAudit {transmission_id} for enterprise enrollment {enrollment}, '
            'email {degreed_user_email}, and course {course_id}>'.format(
                transmission_id=self.id,
                enrollment=self.enterprise_course_enrollment_id,
                degreed_user_email=self.degreed_user_email,
                course_id=self.course_id
            )
        )

    def __repr__(self):
        """
        Return uniquely identifying string representation.
        """
        return self.__str__()



# -*- coding: utf-8 -*-
"""
Enterprise Integrated Channel Blackboard Django application initialization.
"""

from django.apps import AppConfig

CHANNEL_NAME = 'integrated_channels.blackboard'
VERBOSE_NAME = 'Enterprise Blackboard Integration'
BRIEF_CHANNEL_NAME = 'blackboard'


class BlackboardConfig(AppConfig):
    """
    Configuration for the Enterprise Integrated Channel Blackboard Django application.
    """
    name = CHANNEL_NAME
    verbose_name = VERBOSE_NAME
    oauth_token_auth_path = "learn/api/public/v1/oauth2/token"
    brief_channel_name = BRIEF_CHANNEL_NAME



# -*- coding: utf-8 -*-
"""
Learner data exporter for Enterprise Integrated Channel Moodle.
"""

from logging import getLogger

from django.apps import apps

from integrated_channels.catalog_service_utils import get_course_id_for_enrollment
from integrated_channels.integrated_channel.exporters.learner_data import LearnerExporter
from integrated_channels.utils import generate_formatted_log, parse_datetime_to_epoch_millis

LOGGER = getLogger(__name__)


class MoodleLearnerExporter(LearnerExporter):
    """
    Class to provide a Moodle learner data transmission audit prepared for serialization.
    """

    def get_learner_data_records(
            self,
            enterprise_enrollment,
            completed_date=None,
            course_completed=False,
            **kwargs
    ):  # pylint: disable=arguments-differ
        """
        Return a MoodleLearnerDataTransmissionAudit with the given enrollment and course completion data.
        If no remote ID can be found, return None.
        """
        enterprise_learner = enterprise_enrollment.enterprise_customer_user
        completed_timestamp = None
        if completed_date is not None:
            completed_timestamp = parse_datetime_to_epoch_millis(completed_date)

        if enterprise_learner.user_email is None:
            LOGGER.debug(generate_formatted_log(
                'moodle',
                enterprise_learner.enterprise_customer.uuid,
                enterprise_learner.user_id,
                None,
                ('get_learner_data_records finished. No learner data was sent for this LMS User Id because '
                 'Moodle User ID not found for [{name}]'.format(
                     name=enterprise_enrollment.enterprise_customer_user.enterprise_customer.name
                 ))))
            return None

        MoodleLearnerDataTransmissionAudit = apps.get_model(
            'moodle',
            'MoodleLearnerDataTransmissionAudit'
        )

        percent_grade = kwargs.get('grade_percent', None)

        return [
            MoodleLearnerDataTransmissionAudit(
                enterprise_course_enrollment_id=enterprise_enrollment.id,
                moodle_user_email=enterprise_learner.user_email,
                course_id=get_course_id_for_enrollment(enterprise_enrollment),
                course_completed=course_completed,
                grade=percent_grade,
                completed_timestamp=completed_timestamp,
            ),
            MoodleLearnerDataTransmissionAudit(
                enterprise_course_enrollment_id=enterprise_enrollment.id,
                moodle_user_email=enterprise_learner.user_email,
                course_id=enterprise_enrollment.course_id,
                course_completed=course_completed,
                grade=percent_grade,
                completed_timestamp=completed_timestamp,
            )
        ]



"""
Views containing APIs for Canvas integrated channel
"""

import requests
from rest_framework import generics
from rest_framework.exceptions import APIException, NotFound, ParseError
from rest_framework.response import Response
from six.moves.urllib.parse import urljoin

from django.apps import apps
from django.conf import settings
from django.core.exceptions import ValidationError

from enterprise.utils import get_enterprise_customer
from integrated_channels.canvas.models import CanvasEnterpriseCustomerConfiguration


class CanvasCompleteOAuthView(generics.ListAPIView):
    """
        **Use Cases**

            Retrieve and save a Canvas OAuth refresh token after an enterprise customer
            authorizes to integrated courses.

        **Example Requests**

            GET /canvas/oauth-complete?code=123abc&state=abc123

        **Query Parameters for GET**

            * code: The one time use string generated by the Canvas API used to fetch the
            access and refresh tokens for integrating with Canvas.

            * state: The user's enterprise customer uuid used to associate the incoming
            code with an enterprise configuration model.

        **Response Values**

            If the request for information about the course list is successful, an HTTP 200 "OK" response
            is returned.

            If code is not provided, a 400 error is returned.

            If state is not provided, a 400 error is returned.

            If the specified state is not valid or contained in the set of registered enterprise customers
            a 404 error is returned.

    """
    def get(self, request, *args, **kwargs):
        app_config = apps.get_app_config('canvas')
        canvas_oauth_token_path = app_config.oauth_token_auth_path

        # Check if Canvas encountered an error when generating the oauth code.
        canvas_request_error = request.GET.get('error')
        if canvas_request_error:
            raise APIException(
                'Canvas OAuth API encountered an error when generating client code- error: {} description: {}'.format(
                    canvas_request_error,
                    request.GET.get('error_description')
                )
            )

        # Retrieve the newly generated code and state (Enterprise user's ID)
        client_code = request.GET.get('code')
        state_uuid = request.GET.get('state')

        if not state_uuid:
            raise ParseError("Canvas Configuration uuid required to integrate with Canvas.")

        if not client_code:
            raise ParseError("Client code required to integrate with Canvas.")

        try:
            enterprise_config = CanvasEnterpriseCustomerConfiguration.objects.get(uuid=state_uuid)
        except (CanvasEnterpriseCustomerConfiguration.DoesNotExist, ValidationError):
            enterprise_config = None

        # old urls may use the enterprise customer uuid in place of the config uuid, so lets fallback
        if not enterprise_config:
            enterprise_customer = get_enterprise_customer(state_uuid)

            if not enterprise_customer:
                raise NotFound(f"No state data found for given uuid: {state_uuid}.")

            try:
                enterprise_config = CanvasEnterpriseCustomerConfiguration.objects.get(
                    enterprise_customer=enterprise_customer
                )
            except CanvasEnterpriseCustomerConfiguration.DoesNotExist as error:
                raise NotFound(
                    f"No Canvas configuration found for state: {state_uuid}"
                ) from error

        access_token_request_params = {
            'grant_type': 'authorization_code',
            'client_id': enterprise_config.client_id,
            'client_secret': enterprise_config.client_secret,
            'redirect_uri': settings.LMS_INTERNAL_ROOT_URL + "/canvas/oauth-complete",
            'code': client_code,
        }

        auth_token_url = urljoin(
            enterprise_config.canvas_base_url,
            canvas_oauth_token_path
        )

        auth_response = requests.post(auth_token_url, access_token_request_params)

        try:
            data = auth_response.json()
            refresh_token = data['refresh_token']
        except (KeyError, ValueError) as error:
            raise requests.RequestException(response=auth_response) from error

        enterprise_config.refresh_token = refresh_token
        enterprise_config.save()

        return Response()



"""
Viewsets for integrated_channels/v1/sap_success_factors/
"""
from rest_framework import exceptions, permissions, status, viewsets

from integrated_channels.api.v1.mixins import PermissionRequiredForIntegratedChannelMixin
from integrated_channels.sap_success_factors.models import SAPSuccessFactorsEnterpriseCustomerConfiguration

from .serializers import SAPSuccessFactorsConfigSerializer


class SAPSuccessFactorsConfigurationViewSet(PermissionRequiredForIntegratedChannelMixin, viewsets.ModelViewSet):
    serializer_class = SAPSuccessFactorsConfigSerializer
    permission_classes = (permissions.IsAuthenticated,)
    permission_required = 'enterprise.can_access_admin_dashboard'

    configuration_model = SAPSuccessFactorsEnterpriseCustomerConfiguration



# -*- coding: utf-8 -*-
"""
The Cornerstone Integrated Channel package.
"""

__version__ = "0.1.0"

default_app_config = ("integrated_channels.cornerstone.apps."
                      "CornerstoneConfig")



# -*- coding: utf-8 -*-
"""
Filters for enterprise API.
"""

from rest_framework import filters
from rest_framework.exceptions import ValidationError

from django.contrib import auth

from enterprise.models import EnterpriseCustomerUser, SystemWideEnterpriseUserRoleAssignment

User = auth.get_user_model()


class UserFilterBackend(filters.BaseFilterBackend):
    """
    Filter backend for any view that needs to filter against the requesting user's ID.

    * Staff users will bypass this filter.
    * Non-staff users will receive only those objects that match their own user ID.

    This requires that `USER_ID_FILTER` be set in the view as a class variable, to identify
    the object's relationship to a user ID.
    """

    def filter_queryset(self, request, queryset, view):
        """
        Filter only for the user's ID if non-staff.
        """
        if not request.user.is_staff:
            filter_kwargs = {view.USER_ID_FILTER: request.user.id}
            queryset = queryset.filter(**filter_kwargs)
        return queryset


class EnterpriseCustomerUserFilterBackend(filters.BaseFilterBackend):
    """
    Allow filtering on the enterprise customer user api endpoint.
    """

    def _filter_by_user_attributes(self, request, queryset):
        """
        Filter queryset by email or username.
        """
        email = request.query_params.get('email', None)
        username = request.query_params.get('username', None)

        query_params = {}

        if email:
            query_params.update(email=email)
        if username:
            query_params.update(username=username)

        if query_params:
            users = User.objects.filter(**query_params).values_list('id', flat=True)
            return queryset.filter(user_id__in=users)

        return queryset

    def _filter_by_enterprise_attributes(self, request, queryset):
        """
        Filter queryset by enterprise_customer_uuid or enterprise role.
        """
        enterprise_customer_uuid = request.query_params.get('enterprise_customer_uuid', None)
        role = request.query_params.get('role', None)

        query_params = {}

        if enterprise_customer_uuid:
            query_params.update(enterprise_customer_id=enterprise_customer_uuid)

        if role:
            role_assignment_filters = {'role__name': role}

            if not enterprise_customer_uuid:
                raise ValidationError('Cannot filter by role without providing enterprise_customer_uuid.')

            role_assignment_filters.update(enterprise_customer_id=enterprise_customer_uuid)

            users_with_role = SystemWideEnterpriseUserRoleAssignment.objects.filter(
                **role_assignment_filters
            ).values_list('user_id', flat=True)

            query_params.update(user_id__in=users_with_role)

        if query_params:
            return queryset.filter(**query_params)

        return queryset

    def filter_queryset(self, request, queryset, view):
        """
        Apply incoming filters only if user is staff. If not, only filter by user's ID.
        """
        if request.user.is_staff:
            queryset = self._filter_by_user_attributes(request, queryset)
            queryset = self._filter_by_enterprise_attributes(request, queryset)
        else:
            queryset = queryset.filter(user_id=request.user.id)

        return queryset


class EnterpriseLinkedUserFilterBackend(filters.BaseFilterBackend):
    """
    Filter backend to return user's linked enterprises only

    * Staff users will bypass this filter.
    * Non-staff users will receive only their linked enterprises.
    """

    def filter_queryset(self, request, queryset, view):
        """
        Filter out enterprise customer if learner is not linked
        """
        if not request.user.is_staff:
            filter_kwargs = {
                view.USER_ID_FILTER: request.user.id,
                'enterprise_customer_users__linked': 1
            }
            queryset = queryset.filter(**filter_kwargs)

        return queryset


class EnterpriseCustomerInviteKeyFilterBackend(filters.BaseFilterBackend):
    """
    Filter backend to return invite keys under the user's enterprise(s) only.
    Supports filtering by enterprise_customer_uuid.

    * Staff users will bypass this filter.
    """

    def filter_queryset(self, request, queryset, view):
        query_params = {}

        if not request.user.is_staff:
            user_enterprise_customer_uuids = [
                ec.enterprise_customer_id for ec in EnterpriseCustomerUser.objects.filter(user_id=request.user.id)
            ]
            query_params.update(enterprise_customer_id__in=user_enterprise_customer_uuids)

        enterprise_customer_uuid = request.query_params.get('enterprise_customer_uuid', None)
        if enterprise_customer_uuid:
            query_params.update(enterprise_customer_id=enterprise_customer_uuid)

        return queryset.filter(**query_params)



# -*- coding: utf-8 -*-
"""
Utility functions for the SAP SuccessFactors integrated channel.
"""

from logging import getLogger

from integrated_channels.sap_success_factors.constants import SUCCESSFACTORS_OCN_LANGUAGE_CODES

LOGGER = getLogger(__name__)


def transform_language_code(code):
    """
    Transform ISO language code (e.g. en-us) to the language name expected by SAPSF.
    """
    if code is None:
        return 'English'

    components = code.split('-', 2)
    language_code = components[0]
    try:
        country_code = components[1]
    except IndexError:
        country_code = '_'

    language_family = SUCCESSFACTORS_OCN_LANGUAGE_CODES.get(language_code)
    if not language_family:
        return 'English'

    return language_family.get(country_code, language_family['_'])



# -*- coding: utf-8 -*-
"""
Django management command to unlink the learners, Delete the enrollments and remove the DSC.
"""
import csv
import logging

from edx_django_utils.cache import TieredCache, get_cache_key

from django.contrib import auth
from django.core.management import BaseCommand

from enterprise.models import EnterpriseCustomer, EnterpriseCustomerUser, PendingEnterpriseCustomerUser

LOGGER = logging.getLogger(__name__)

User = auth.get_user_model()


class Command(BaseCommand):
    """
    Django management command to unlink the learners, Delete the enrollments and remove the DSC

    Example usage:
    ./manage.py lms unlink_enterprise_customer_learners -e <enterprise-uuid> --data-csv /path/file.csv
    ./manage.py lms unlink_enterprise_customer_learners -e <enterprise-uuid> --data-csv /path/file.csv --skip-unlink

    """

    def add_arguments(self, parser):
        parser.add_argument(
            '--data-csv',
            action='store',
            dest='data_csv',
            help='Path of csv to read emails to be unlinked.',
            type=str,
        )

        parser.add_argument(
            '-e',
            '--enterprise_customer_uuid',
            action='store',
            dest='enterprise_customer_uuid',
            help='Run this command for only the given EnterpriseCustomer UUID.'
        )

        parser.add_argument(
            '--skip-unlink',
            action='store_true',
            dest='skip_unlink',
            default=False,
            help='Specify to remove the DSC without unlinking learner for given data_csv.'
        )

    def handle(self, *args, **options):
        csv_path = options['data_csv']
        skip_unlink = options['skip_unlink']
        enterprise_customer_uuid = options['enterprise_customer_uuid']
        enterprise_customer = EnterpriseCustomer.objects.get(uuid=enterprise_customer_uuid)

        LOGGER.info('[Unlink and Remove DSC]  Process started.')

        results = {}

        with open(csv_path) as csv_file:
            rows = list(csv.DictReader(csv_file))
        csv_file.close()

        for row in rows:
            email = row['email']
            results[email] = {'Unlinked': None, 'Removed_DSC': None}

            if not skip_unlink:
                # Unlink the user.
                try:
                    EnterpriseCustomerUser.objects.unlink_user(
                        enterprise_customer=enterprise_customer, user_email=email
                    )
                    results[email]['Unlinked'] = 'Success'
                except (EnterpriseCustomerUser.DoesNotExist, PendingEnterpriseCustomerUser.DoesNotExist):
                    message = 'Email {email} is not associated with Enterprise Customer {ec_name}'.format(
                        email=email, ec_name=enterprise_customer.name
                    )
                    results[email]['Unlinked'] = message

            # Deleting the DSC record.
            try:
                user = User.objects.get(email=email)
                enterprise_customer_user = EnterpriseCustomerUser.all_objects.get(
                    enterprise_customer__uuid=enterprise_customer_uuid,
                    user_id=user.id
                )
                data_sharing_consent_records = enterprise_customer_user.data_sharing_consent_records
                course_ids = list(data_sharing_consent_records.values_list('course_id', flat=True))
                data_sharing_consent_records.delete()
                # Deleting the DCS cache
                for course_id in course_ids:
                    consent_cache_key = get_cache_key(
                        type='data_sharing_consent_needed',
                        user_id=user.id,
                        course_id=course_id
                    )
                    TieredCache.delete_all_tiers(consent_cache_key)
                results[email]['Removed_DSC'] = course_ids
            except (User.DoesNotExist, EnterpriseCustomerUser.DoesNotExist) as excep:
                results[email]['Removed_DSC'] = str(excep)

        LOGGER.info('[Unlink and Remove DSC] Execution completed for enterprise: %s, \nResults: %s',
                    enterprise_customer_uuid, results)



# -*- coding: utf-8 -*-
"""
Helper functions for the Consent application.
"""

from django.apps import apps

from consent.models import ProxyDataSharingConsent
from enterprise.api_client.discovery import get_course_catalog_api_service_client
from enterprise.utils import get_enterprise_customer


def get_data_sharing_consent(username, enterprise_customer_uuid, course_id=None, program_uuid=None):
    """
    Get the data sharing consent object associated with a certain user, enterprise customer, and other scope.

    :param username: The user that grants consent
    :param enterprise_customer_uuid: The consent requester
    :param course_id (optional): A course ID to which consent may be related
    :param program_uuid (optional): A program to which consent may be related
    :return: The data sharing consent object, or None if the enterprise customer for the given UUID does not exist.
    """
    EnterpriseCustomer = apps.get_model('enterprise', 'EnterpriseCustomer')
    try:
        if course_id:
            return get_course_data_sharing_consent(username, course_id, enterprise_customer_uuid)
        return get_program_data_sharing_consent(username, program_uuid, enterprise_customer_uuid)
    except EnterpriseCustomer.DoesNotExist:
        return None


def get_course_data_sharing_consent(username, course_id, enterprise_customer_uuid):
    """
    Get the data sharing consent object associated with a certain user of a customer for a course.

    :param username: The user that grants consent.
    :param course_id: The course for which consent is granted.
    :param enterprise_customer_uuid: The consent requester.
    :return: The data sharing consent object
    """
    # Prevent circular imports.
    DataSharingConsent = apps.get_model('consent', 'DataSharingConsent')
    return DataSharingConsent.objects.proxied_get(
        username=username,
        course_id=course_id,
        enterprise_customer__uuid=enterprise_customer_uuid
    )


def get_program_data_sharing_consent(username, program_uuid, enterprise_customer_uuid):
    """
    Get the data sharing consent object associated with a certain user of a customer for a program.

    :param username: The user that grants consent.
    :param program_uuid: The program for which consent is granted.
    :param enterprise_customer_uuid: The consent requester.
    :return: The data sharing consent object
    """
    enterprise_customer = get_enterprise_customer(enterprise_customer_uuid)
    discovery_client = get_course_catalog_api_service_client(enterprise_customer.site)
    course_ids = discovery_client.get_program_course_keys(program_uuid)
    child_consents = (
        get_data_sharing_consent(username, enterprise_customer_uuid, course_id=individual_course_id)
        for individual_course_id in course_ids
    )
    return ProxyDataSharingConsent.from_children(program_uuid, *child_consents)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# pylint: disable=C0111,W6005,W6100


import os
import re
import sys

from setuptools import setup

VCS_PREFIXES = ('git+', 'hg+', 'bzr+', 'svn+', '-e git+')


def get_version(*file_paths):
    """
    Extract the version string from the file at the given relative path fragments.
    """
    filename = os.path.join(os.path.dirname(__file__), *file_paths)
    version_file = open(filename).read()
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


def get_requirements(requirements_file):
    """
    Get the contents of a file listing the requirements
    """
    lines = open(requirements_file).readlines()
    dependencies = []
    dependency_links = []

    for line in lines:
        package = line.strip()
        if package.startswith('#'):
            # Skip pure comment lines
            continue

        package, __, __ = package.partition(' #')
        package = package.strip()

        if any(package.startswith(prefix) for prefix in VCS_PREFIXES):
            # VCS reference for dev purposes, expect a trailing comment
            # with the normal requirement
            package_link, __, package = package.rpartition('#')

            # Remove -e <version_control> string
            package_link = re.sub(r'(.*)(?P<dependency_link>https?.*$)', r'\g<dependency_link>', package_link)
            package = re.sub(r'(egg=)?(?P<package_name>.*)==.*$', r'\g<package_name>', package)
            package_version = re.sub(r'.*[^=]==', '', line.strip())

            if package:
                dependency_links.append(
                    '{package_link}#egg={package}-{package_version}'.format(
                        package_link=package_link,
                        package=package,
                        package_version=package_version,
                    )
                )
        else:
            # Ignore any trailing comment
            package, __, __ = package.partition('#')
            # Remove any whitespace and assume non-empty results are dependencies
            package = package.strip()

        if package:
            dependencies.append(package)
    return dependencies, dependency_links


VERSION = get_version("enterprise", "__init__.py")

if sys.argv[-1] == "tag":
    print("Tagging the version on github:")
    os.system("git tag -a %s -m 'version %s'" % (VERSION, VERSION))
    os.system("git push --tags")
    sys.exit()

base_path = os.path.dirname(__file__)

README = open(os.path.join(base_path, "README.rst")).read()
CHANGELOG = open(os.path.join(base_path, "CHANGELOG.rst")).read()
REQUIREMENTS, DEPENDENCY_LINKS = get_requirements(os.path.join(base_path, 'requirements', 'base.in'))

setup(
    name="edx-enterprise",
    version=VERSION,
    description="""Your project description goes here""",
    long_description=README + "\n\n" + CHANGELOG,
    author="edX",
    author_email="oscm@edx.org",
    url="https://github.com/edx/edx-enterprise",
    packages=[
        "enterprise",
        "consent",
        "integrated_channels",
        "integrated_channels.integrated_channel",
        "integrated_channels.degreed",
        "integrated_channels.degreed2",
        "integrated_channels.canvas",
        "integrated_channels.blackboard",
        "integrated_channels.cornerstone",
        "integrated_channels.moodle",
        "integrated_channels.sap_success_factors",
        "integrated_channels.xapi",
        "enterprise_learner_portal",
    ],
    include_package_data=True,
    install_requires=REQUIREMENTS,
    dependency_links=DEPENDENCY_LINKS,
    license="AGPL 3.0",
    zip_safe=False,
    keywords="Django edx",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Framework :: Django",
        "Framework :: Django :: 2.2",
        "Framework :: Django :: 3.0",
        "Framework :: Django :: 3.1",
        "Framework :: Django :: 3.2",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
    ],
)



# -*- coding: utf-8 -*-
"""
Learner data exporter for Enterprise Integrated Channel Cornerstone.
"""

from logging import getLogger

from django.apps import apps

from integrated_channels.catalog_service_utils import get_course_id_for_enrollment
from integrated_channels.cornerstone.utils import get_or_create_key_pair
from integrated_channels.integrated_channel.exporters.learner_data import LearnerExporter
from integrated_channels.utils import generate_formatted_log

LOGGER = getLogger(__name__)


class CornerstoneLearnerExporter(LearnerExporter):
    """
    Class to provide a Cornerstone learner data transmission audit prepared for serialization.
    """

    def get_learner_data_records(
            self,
            enterprise_enrollment,
            completed_date=None,
            grade=None,
            course_completed=False,
            **kwargs,
    ):  # pylint: disable=arguments-differ
        """
        Return a CornerstoneLearnerDataTransmissionAudit with the given enrollment and course completion data.

        CornerstoneLearnerDataTransmissionAudit object should exit already if not then return None.
        """

        CornerstoneLearnerDataTransmissionAudit = apps.get_model(
            'cornerstone',
            'CornerstoneLearnerDataTransmissionAudit'
        )

        try:
            course_id = get_course_id_for_enrollment(enterprise_enrollment)
            key_mapping = get_or_create_key_pair(course_id)
            csod_learner_data_transmission = CornerstoneLearnerDataTransmissionAudit.objects.get(
                user_id=enterprise_enrollment.enterprise_customer_user.user.id,
                course_id=key_mapping.external_course_id,
            )
            csod_learner_data_transmission.enterprise_course_enrollment_id = enterprise_enrollment.id
            csod_learner_data_transmission.grade = grade
            csod_learner_data_transmission.course_completed = course_completed
            csod_learner_data_transmission.completed_timestamp = completed_date
            return [
                csod_learner_data_transmission
            ]
        except CornerstoneLearnerDataTransmissionAudit.DoesNotExist:
            LOGGER.info(generate_formatted_log(
                self.enterprise_configuration.channel_code(),
                enterprise_enrollment.enterprise_customer_user.enterprise_customer.uuid,
                enterprise_enrollment.enterprise_customer_user.user_id,
                None,
                ('get_learner_data_records finished. No learner data was sent for this LMS User Id because '
                 'Cornerstone User ID not found for [{name}]'.format(
                     name=enterprise_enrollment.enterprise_customer_user.enterprise_customer.name
                 ))))
            return None



# -*- coding: utf-8 -*-

"""
Statements base for xAPI.
"""

from tincan import Activity, ActivityDefinition, Agent, LanguageMap, Statement

from integrated_channels.xapi.constants import X_API_ACTIVITY_COURSE


class EnterpriseStatement(Statement):
    """
    Base statement for enterprise events.
    """

    def _get_actor_name(self, user, user_social_auth):
        """
        Returns the name of the actor based on provided information and defined rules

        Arguments:
            user (User): User.
            user_social_auth (UserSocialAuth): UserSocialAuth.
        """
        social_auth_uid = user_social_auth.uid if user_social_auth else ''
        sso_id = social_auth_uid.split(':')[-1]
        actor_name = sso_id if sso_id else user.email
        return actor_name

    def get_actor(self, user, user_social_auth):
        """
        Returns the actor component of the Enterprise xAPI statement.
        Arguments:
            user (User): User.
            user_social_auth (UserSocialAuth): UserSocialAuth.
        """
        name = self._get_actor_name(user, user_social_auth)
        return Agent(
            name=name,
            mbox=u'mailto:{email}'.format(email=user.email),
        )

    def get_object(self, domain, course_overview, object_type):
        """
        Returns the object (activity) component of the Enterprise xAPI statement.
        Arguments:
            course_overview (CourseOverview): CourseOverview.
            object_type (string): Object type for activity.
        """
        name = (course_overview.display_name or '').encode("ascii", "ignore").decode('ascii')

        description = (course_overview.short_description or '').encode("ascii", "ignore").decode('ascii')

        activity_id = course_overview.id
        if object_type is not None and object_type == 'course':
            activity_id = course_overview.course_key

        xapi_activity_id = 'https://{domain}/xapi/activities/{object_type}/{activity_id}'.format(
            domain=domain,
            object_type=object_type,
            activity_id=activity_id
        )

        xapi_object_extensions = {}

        course_key_keyname = 'https://{domain}/course/key'.format(domain=domain)
        xapi_object_extensions[course_key_keyname] = course_overview.course_key

        course_uuid_keyname = 'https://{domain}/course/uuid'.format(domain=domain)
        xapi_object_extensions[course_uuid_keyname] = course_overview.course_uuid

        return Activity(
            id=xapi_activity_id,
            definition=ActivityDefinition(
                type=X_API_ACTIVITY_COURSE,
                name=LanguageMap({'en-US': name}),
                description=LanguageMap({'en-US': description}),
                extensions=xapi_object_extensions,
            ),
        )



# Generated by Django 2.2.19 on 2021-08-03 18:54

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise', '0138_bulkcatalogqueryupdatecommandconfiguration'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='adminnotification',
            options={'ordering': ('start_date',), 'verbose_name': 'Admin Notification', 'verbose_name_plural': 'Admin Notifications'},
        ),
        migrations.AddField(
            model_name='adminnotification',
            name='title',
            field=models.CharField(default='New Notification', help_text='Notification banner title which will appear to enterprise admin on admin portal.', max_length=255),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='adminnotification',
            name='text',
            field=models.CharField(help_text='Notification banner text which will appear to enterprise admin on admin portal.', max_length=255),
        ),
    ]



# -*- coding: utf-8 -*-
"""
URL definitions for enterprise_learner_portal API endpoint.
"""

from django.conf.urls import include, url

urlpatterns = [
    url(r'^v1/', include('enterprise_learner_portal.api.v1.urls'), name='v1')
]



# -*- coding: utf-8 -*-
"""
Client for communicating with the Enterprise API.
"""

import json
from collections import OrderedDict
from logging import getLogger

from edx_rest_api_client.client import EdxRestApiClient
from requests.exceptions import ConnectionError, Timeout  # pylint: disable=redefined-builtin
from slumber.exceptions import HttpNotFoundError, SlumberBaseException

from django.conf import settings

from enterprise import utils
from enterprise.api_client.lms import JwtLmsApiClient

LOGGER = getLogger(__name__)


class EnterpriseCatalogApiClient(JwtLmsApiClient):
    """
    Object builds an API client to make calls to the Enterprise Catalog API.
    """

    API_BASE_URL = settings.ENTERPRISE_CATALOG_INTERNAL_ROOT_URL + '/api/v1/'
    ENTERPRISE_CATALOG_ENDPOINT = 'enterprise-catalogs'
    GET_CONTENT_METADATA_ENDPOINT = ENTERPRISE_CATALOG_ENDPOINT + '/{}/get_content_metadata'
    REFRESH_CATALOG_ENDPOINT = ENTERPRISE_CATALOG_ENDPOINT + '/{}/refresh_metadata'
    CATALOG_DIFF_ENDPOINT = ENTERPRISE_CATALOG_ENDPOINT + '/{}/generate_diff'
    ENTERPRISE_CUSTOMER_ENDPOINT = 'enterprise-customer'
    APPEND_SLASH = True
    GET_CONTENT_METADATA_PAGE_SIZE = getattr(settings, 'ENTERPRISE_CATALOG_GET_CONTENT_METADATA_PAGE_SIZE', 50)

    def __init__(self, user=None):
        user = user if user else utils.get_enterprise_worker_user()
        super().__init__(user)

    @JwtLmsApiClient.refresh_token
    def create_enterprise_catalog(
            self,
            catalog_uuid,
            enterprise_id,
            enterprise_name,
            title,
            content_filter,
            enabled_course_modes,
            publish_audit_enrollment_urls,
            catalog_query_uuid,
            query_title=None):
        """Creates an enterprise catalog."""
        endpoint = getattr(self.client, self.ENTERPRISE_CATALOG_ENDPOINT)
        post_data = {
            'uuid': catalog_uuid,
            'title': title,
            'enterprise_customer': enterprise_id,
            'enterprise_customer_name': enterprise_name,
            'content_filter': content_filter,
            'enabled_course_modes': enabled_course_modes,
            'publish_audit_enrollment_urls': json.dumps(publish_audit_enrollment_urls),
            'catalog_query_uuid': catalog_query_uuid,
            'query_title': query_title,
        }
        try:
            LOGGER.info(
                'Creating Enterprise Catalog %s in the Enterprise Catalog Service with params: %s',
                catalog_uuid,
                json.dumps(post_data)
            )
            return endpoint.post(post_data)
        except (SlumberBaseException, ConnectionError, Timeout) as exc:
            LOGGER.exception(
                'Failed to create EnterpriseCustomer Catalog [%s] in enterprise-catalog due to: [%s]',
                catalog_uuid, str(exc)
            )
            return {}

    @JwtLmsApiClient.refresh_token
    def get_enterprise_catalog(self, catalog_uuid, should_raise_exception=True):
        """
        Gets an enterprise catalog.

        Arguments:
            catalog_uuid (uuid): The uuid of an EnterpriseCatalog instance
            should_raise_exception (bool): Whether an exception should be logged if
                a catalog does not yet exist in enterprise-catalog. Defaults to True.

        Returns:
            dict: a dictionary representing an enterprise catalog
        """
        endpoint = getattr(self.client, self.ENTERPRISE_CATALOG_ENDPOINT)(catalog_uuid)
        try:
            return endpoint.get()
        except (SlumberBaseException, ConnectionError, Timeout) as exc:
            if should_raise_exception:
                LOGGER.exception(
                    'Failed to get EnterpriseCustomer Catalog [%s] in enterprise-catalog due to: [%s]',
                    catalog_uuid, str(exc)
                )
            return {}

    @JwtLmsApiClient.refresh_token
    def get_catalog_diff(self, enterprise_customer_catalog, content_keys, should_raise_exception=True):
        """
        Gets the representational difference between a list of course keys and the current state of content under an
        enterprise catalog. This difference is returned as three buckets of data: `items_not_found`,
        `items_not_included` and `items_found`.

        Arguments:
            enterprise_customer_catalog (EnterpriseCustomerCatalog): The catalog object whose content is being diffed.
            content_keys (list): List of string content keys
            should_raise_exception (Bool): Optional param for whether or not api response exceptions should be raised.

        Response:
            items_to_create: [{
               "course_key": <content key>>
            }, {
            ..
            }]

            items_to_delete: [{
               "course_key": <content key>>
            }, {
            ..
            }]

            items_found: [{
                "course_key": <content key>>,
                "date_updated": <content last modified datetime>
            }, {
                ...
            }]
        """
        catalog_uuid = enterprise_customer_catalog.uuid
        endpoint = getattr(self.client, self.CATALOG_DIFF_ENDPOINT.format(catalog_uuid))
        body = {'content_keys': content_keys}

        items_to_delete = {}
        items_to_create = []
        items_found = []
        try:
            response = endpoint.post(body)
            items_to_delete = response.get('items_not_found')
            items_to_create = response.get('items_not_included')
            items_found = response.get('items_found')

        except (SlumberBaseException, ConnectionError, Timeout) as exc:
            LOGGER.exception(
                'Failed to get EnterpriseCustomer Catalog [%s] in enterprise-catalog due to: [%s]',
                catalog_uuid, str(exc)
            )
            if should_raise_exception:
                raise

        return items_to_create, items_to_delete, items_found

    @staticmethod
    def get_content_metadata_url(uuid):
        """
        Get the url for the preview information for an enterprise catalog
        """
        return EnterpriseCatalogApiClient.API_BASE_URL + \
            EnterpriseCatalogApiClient.GET_CONTENT_METADATA_ENDPOINT.format(uuid)

    @JwtLmsApiClient.refresh_token
    def update_enterprise_catalog(self, catalog_uuid, **kwargs):
        """Updates an enterprise catalog."""
        endpoint = getattr(self.client, self.ENTERPRISE_CATALOG_ENDPOINT)(catalog_uuid)
        try:
            LOGGER.info(
                'Updating Enterprise Catalog %s in the Enterprise Catalog Service with params: %s',
                catalog_uuid,
                json.dumps(kwargs)
            )
            return endpoint.put(kwargs)
        except (SlumberBaseException, ConnectionError, Timeout) as exc:
            LOGGER.exception(
                'Failed to update EnterpriseCustomer Catalog [%s] in enterprise-catalog due to: [%s]',
                catalog_uuid, str(exc)
            )
            return {}

    @JwtLmsApiClient.refresh_token
    def delete_enterprise_catalog(self, catalog_uuid):
        """Deletes an enterprise catalog."""
        endpoint = getattr(self.client, self.ENTERPRISE_CATALOG_ENDPOINT)(catalog_uuid)
        try:
            return endpoint.delete()
        except HttpNotFoundError:
            LOGGER.warning(
                'Deleted EnterpriseCustomerCatalog [%s] that was not in enterprise-catalog',
                catalog_uuid
            )
            return {}
        except (SlumberBaseException, ConnectionError, Timeout) as exc:
            LOGGER.exception(
                'Failed to delete EnterpriseCustomer Catalog [%s] in enterprise-catalog due to: [%s]',
                catalog_uuid, str(exc)
            )
            return {}

    def traverse_get_content_metadata(self, endpoint, query, catalog_uuid):
        """
        Helper method to traverse over a paginated response from the enterprise-catalog service's `get_content_metadata`
        endpoint.
        """
        content_metadata = OrderedDict()
        try:
            response = endpoint.get(**query)
            for item in utils.traverse_pagination(response, endpoint):
                content_id = utils.get_content_metadata_item_id(item)
                content_metadata[content_id] = item
        except (SlumberBaseException, ConnectionError, Timeout):
            LOGGER.exception(
                f'Failed to get content metadata for Catalog {catalog_uuid} in enterprise-catalog',
                exc_info=True,
            )
            raise

        return content_metadata

    @JwtLmsApiClient.refresh_token
    def get_content_metadata(self, enterprise_customer, enterprise_catalogs=None, content_keys_filter=None):
        """
        Return all content metadata contained in the catalogs associated with the EnterpriseCustomer.

        Arguments:
            enterprise_customer (EnterpriseCustomer): The EnterpriseCustomer to return content metadata for.
            enterprise_catalogs (EnterpriseCustomerCatalog): Optional list of EnterpriseCustomerCatalog objects.
            content_keys_filter (List): List of content keys to filter by in the content metadata endpoint

        Returns:
            list: List of dicts containing content metadata.
        """

        content_metadata = OrderedDict()
        enterprise_customer_catalogs = enterprise_catalogs or enterprise_customer.enterprise_customer_catalogs.all()
        for enterprise_customer_catalog in enterprise_customer_catalogs:
            catalog_uuid = enterprise_customer_catalog.uuid
            endpoint = getattr(self.client, self.GET_CONTENT_METADATA_ENDPOINT.format(catalog_uuid))
            # If content keys filter exists then chunk up the keys into reasonable request sizes
            if content_keys_filter:
                chunked_keys_filter = utils.batch(
                    content_keys_filter,
                    self.GET_CONTENT_METADATA_PAGE_SIZE
                )
                # A chunk can be larger than the page size so traverse pagination for each individual chunk
                for chunk in chunked_keys_filter:
                    query = {'page_size': self.GET_CONTENT_METADATA_PAGE_SIZE, 'content_keys': chunk}
                    content_metadata.update(self.traverse_get_content_metadata(endpoint, query, catalog_uuid))
            # Traverse pagination for the get all content response without filters
            else:
                query = {'page_size': self.GET_CONTENT_METADATA_PAGE_SIZE}
                content_metadata.update(self.traverse_get_content_metadata(endpoint, query, catalog_uuid))

        return list(content_metadata.values())

    @JwtLmsApiClient.refresh_token
    def refresh_catalogs(self, enterprise_catalogs):
        """
        Kicks off async tasks to refresh catalogs so recent changes will populate to production without needing to wait
        for the daily jobs to run

        Arguments:
            enterprise_catalogs (EnterpriseCustomerCatalog): List of EnterpriseCustomerCatalog objects to refresh

        Returns:
            Dict: Dict of async task ids for each catalog id
        """
        refreshed_catalogs = {}
        failed_to_refresh_catalogs = []
        for enterprise_customer_catalog in enterprise_catalogs:
            catalog_uuid = enterprise_customer_catalog.uuid
            endpoint = getattr(self.client, self.REFRESH_CATALOG_ENDPOINT.format(catalog_uuid))
            try:
                response = endpoint.post()
                refreshed_catalogs[catalog_uuid] = response['async_task_id']
            except (SlumberBaseException, ConnectionError, Timeout) as exc:
                LOGGER.exception(
                    'Failed to refresh catalog data for catalog [%s] in enterprise-catalog due to: [%s]',
                    catalog_uuid, str(exc)
                )
                failed_to_refresh_catalogs.append(catalog_uuid)

        return refreshed_catalogs, failed_to_refresh_catalogs

    @JwtLmsApiClient.refresh_token
    def contains_content_items(self, catalog_uuid, content_ids):
        """
        Checks whether an enterprise catalog contains the given content

        The enterprise catalog endpoint does not differentiate between course_run_ids and program_uuids so they can
        be used interchangeably. The two query parameters are left in for backwards compatability with edx-enterprise.
        """
        query_params = {'course_run_ids': content_ids}
        endpoint = getattr(self.client, self.ENTERPRISE_CATALOG_ENDPOINT)(catalog_uuid)
        return endpoint.contains_content_items.get(**query_params)['contains_content_items']

    @JwtLmsApiClient.refresh_token
    def enterprise_contains_content_items(self, enterprise_uuid, content_ids):
        """
        Checks whether an enterprise customer has any catalogs that contain the provided content ids.

        The endpoint does not differentiate between course_run_ids and program_uuids so they can be used
        interchangeably. The two query parameters are left in for backwards compatability with edx-enterprise.
        """
        query_params = {'course_run_ids': content_ids}
        endpoint = getattr(self.client, self.ENTERPRISE_CUSTOMER_ENDPOINT)(enterprise_uuid)
        return endpoint.contains_content_items.get(**query_params)['contains_content_items']


class NoAuthEnterpriseCatalogClient:
    """
    Class to build a enterprise catalog client to make calls to the discovery service.
    """

    API_BASE_URL = settings.ENTERPRISE_CATALOG_INTERNAL_ROOT_URL
    APPEND_SLASH = False

    def __init__(self):
        """
        Create an enterprise catalog client.
        """
        self.client = EdxRestApiClient(self.API_BASE_URL, append_slash=self.APPEND_SLASH)

    def get_health(self):
        """
        Retrieve health details for enterprise catalog service.

        Returns:
            dict: Response containing enterprise catalog service health.
        """
        return self.client.health.get()
